# Designed May. 2011 by Pierreden
# Permission to use, copy, modify this software for
# any non-commercial purpose is hereby granted

module Pdn
  module LiveIvy
    module Parameters
      def self.initialize_default
        hsh = {
            :branch   => {
                :nodesize    => 3,
                :pattern     => 0.8,
                :probability => 0.97,
                :radius      => 0.8,
                :floatlength => 150,
                :detail      => 4,
                :spread      => 0

            },
            :leaf     => {
                :pattern     => 1,
                :probability => 0.7,
                :size        => 5
            },
            :gravity  => {
                :weight => 1,
                :vector => [0, 0, -1]
            },
            :wind     => {
                :weight => 0.1
            },
            :random   => {
                :weight => 0.3
            },
            :adhesion =>{
                :max_adhesion_distance => 50,
                :weight                => 0.2
            },
            :legacy   => {
                :weight => 0.8
            }

        }
      end

      def self.set_param(key, val)
        keys = key.split("_")
        hsh  = read_params

        if keys.length == 2
          hsh[keys[0].to_sym][keys[1].to_sym] = val.to_f
        else
          hsh[keys[0].to_sym] = val.to_f
        end
        return write_params(hsh)
      end

      def self.read_params
        str = Sketchup.read_default("LiveIvy", "parameters")
        str = initialize_default.inspect if str.nil?
        return eval(str)
      end

      def self.write_params(val)
        str = val.inspect
        Sketchup.write_default("LiveIvy", "parameters", str)
        return val
      end

      def self.to_array
        array = []
        for key, val in read_params
          if not val.class == Hash
            array.push(["#{key.to_s}", val])
          else
            for key2, val2 in val
              array.push(["#{key.to_s}_#{key2.to_s.gsub("_","")}", val2])
            end
          end
        end
        array.inspect
      end

      def self.to_json
        hsh = read_params
        arr = []
        arr.push(["legacy_weight", hsh[:legacy][:weight]])
        arr.push(["gravity_weight", hsh[:gravity][:weight]])
        arr.push(["random_weight", hsh[:random][:weight]])
        arr.push(["adhesion_weight", hsh[:adhesion][:weight]])
        arr.push(["wind_weight", hsh[:wind][:weight]])
        arr.push(["scene_size", Sketchup.active_model.bounds.diagonal/2])
        arr.push(["branch_nodesize", hsh[:branch][:nodesize]])
        arr.push(["branch_floatlength", hsh[:branch][:floatlength]])
        arr.push(["branch_radius", hsh[:branch][:radius]])
        arr.push(["branch_pattern", hsh[:branch][:pattern]])
        arr.push(["branch_probability", hsh[:branch][:probability]])
        arr.push(["leaf_size", hsh[:leaf][:size]])

        return arr.inspect
      end
    end
    module Definitions
      def self.present?(val)
        model       = Sketchup.active_model
        definitions = model.definitions
        definitions.each do |d|
          return true if d.name == val
        end
        return false
      end
    end
  end
  module Vector3d
    def self.*(v1, f)
      if f.is_a? Numeric
        return Geom::Vector3d.new(v1.x*f.to_f, v1.y*f.to_f, v1.z*f.to_f)
      else
        return Geom::Vector3d.new(v1.x*f.x, v1.y*f.y, v1.z*f.z)
      end
    end

    def self.radial_distance(v)
      Math::sqrt(v.x**2+v.y**2+v.z**2)
    end

    def self.to_polar(v)
      v     = v + Geom::Vector3d.new(0.000001, 0.000001, 0.000001)
      r     = radial_distance(v)
      #elevation
      alpha = Math::acos(v.z/r)
      #Horizontal
      beta  = Math::atan2(v.y, v.x)
      [r, alpha, beta]
    end

    def self.phi_between(v1, v2)
      Pdn::Vector3d.to_polar(v2)[2]-Pdn::Vector3d.to_polar(v1)[2]
    end

    def self.theta_between(v1, v2)
      Pdn::Vector3d.to_polar(v2)[1]-Pdn::Vector3d.to_polar(v1)[1]
    end
  end
  module LiveVector
    module Adhesive
      def facingFaces
        #Return Faces from set which are facing the point

        arr = seed.faces.select do |f|
          nq = @origin_node.position_point.vector_to(f.vertices[0]).normalize.dot(f.normal)
          nq <= 0.0
        end
      end

      def barycentric_faces
        #Return Faces which can adhes
        #Arguments
        #point as Point3d, Faces as Array, maxdistance as float
        adhesive    = []
        point       = @origin_node.position_point
        maxdistance = seed.params[:adhesion][:max_adhesion_distance]
        for face in facingFaces
          p0       = point.project_to_plane(face.plane)
          distance = p0.distance(point)
          if distance < maxdistance
            cP = face.classify_point(p0)
            if cP == 1
              adhesive.push([face, p0, distance])
            end
          end
        end
        #Array of [face as Face, barycentricPoint as Point3d, distance as Float]
        return adhesive
      end

      def barycentric_face
        #Return Faces which can adhes
        #Arguments
        #point as Point3d, Faces as Array, maxdistance as float
        adhesive    = nil
        point       = @origin_node.position_point
        maxdistance = seed.params[:adhesion][:max_adhesion_distance]
        for face in facingFaces
          p0       = point.project_to_plane(face.plane)
          distance = p0.distance(point)
          if distance < maxdistance
            cP = face.classify_point(p0)
            if cP == 1
              maxdistance = distance
              adhesive    = [face, p0, distance]
            end
          end
        end
        #Array of [face as Face, barycentricPoint as Point3d, distance as Float]
        return adhesive
      end

      def calculate_adhesion
        # Calculates the adhesion Vector of a given point in the scene
        # Arguments: point as Point3d, maxDistance as float, faces as Array
        @adhesion_vector = Geom::Vector3d.new([0, 0, 0])
        mindistance      = seed.params[:adhesion][:max_adhesion_distance]
        face             = barycentric_face
        if face && face[1].class == Geom::Point3d
          aVector          = (@origin_node.position_point.vector_to(face[1])).normalize #+Geom::Vector3d.new(0.000001, 0.000001, 0.000001)
          aVector          = Pdn::Vector3d.*(aVector, 1.0-face[2]/seed.params[:adhesion][:max_adhesion_distance])
          #Return Vector3d
          @adhesion_vector = aVector
        end
        return @adhesion_vector
      end

      def calculate_adhesion2
        # Calculates the adhesion Vector of a given point in the scene
        # Arguments: point as Point3d, maxDistance as float, faces as Array

        aVector     = Geom::Vector3d.new([0, 0, 0])
        mindistance = seed.params[:adhesion][:max_adhesion_distance]
        for face, barypoint, distance in barycentric_faces
          if (distance < mindistance)
            mindistance = distance
            aVector     = (@origin_node.position_point.vector_to(barypoint)).normalize #+Geom::Vector3d.new(0.000001, 0.000001, 0.000001)
            aVector     = Pdn::Vector3d.*(aVector, 1.0-distance/seed.params[:adhesion][:max_adhesion_distance])
          end
        end
        #Return Vector3d
        @adhesion_vector = aVector
        return @adhesion_vector
      end

      def adhesion_vector
        @adhesion_vector
      end
    end
    module Collision
      def calculate_collision(temp_vector)
        #Calculates Collision Point
        #pointA as Point3d, pointB as Point3d, extendDistance as Float
        pointA    = @origin_node.position_point
        pointB    = pointA.offset(temp_vector)
        vector    = temp_vector
        distanceA = pointA.distance(pointB)
        intersect = Sketchup.active_model.raytest([pointA, vector])

        if intersect
          intersectionDistance = pointA.distance(intersect[0])
          if intersect[1].last.class == Sketchup::Face

            if intersectionDistance < distanceA
              #Array [IntersectionPoint as Point, IntersectionFaceNormal as Vector3d, ]
              return [pointB.vector_to(intersect[0]), intersect[1].last.normal]
            end
          elsif intersect[1].last.class == Sketchup::Edge
            if intersectionDistance < distanceA
              face = intersect[1].last.faces
              return [pointB.vector_to(intersect[0]), Pdn::Vector3d.*(vector.normalize, -1)] if face
              return [pointB.vector_to(intersect[0]), vector.normalize]
            end
          end
        end
        return nil
      end

      def collision_adjustment_vector(temp_vector)
        @is_climbing = false
        intersection = calculate_collision(temp_vector)
        if intersection
          @is_climbing     = true
          @climbing_normal = intersection[1]
          return @collision_adjustment_vector = intersection[0]+Pdn::Vector3d.*(intersection[1], seed.params[:branch][:radius]/2)
        else
          return [0, 0, 0]
        end
      end
    end
    module Wind
      def wind_vector
        Geom::Vector3d.new(seed.params['wind']['wind_vector'])
      end
    end
    module Random
      def random_vector
        @random_vector = @random_vector || (Geom::Vector3d.new([rand-0.5, rand-0.5, rand-0.5])+Geom::Vector3d.new([0, 0, 0.2]))
      end
    end
    module Gravity
      def gravity_vector
        Geom::Vector3d.new([0, 0, -1])
      end
    end
    module Legacy
      def legacy_vector
        if @origin_node.is_climbing
          vector = @origin_node.sum_vector
          normal = @origin_node.climbing_normal
          rV     = (vector-normal.multiply(2*vector.dot(normal))).normalize
          pV     = (vector-normal.multiply(1*vector.dot(normal))).normalize
          Geom::linear_combination(rn=rand, pV, 1-rn, rV)
        else
          @origin_node.sum_vector
        end
      end
    end
  end #Impel
  module Entities
    def self.all_faces
      wf = lambda do |c|
        faces = []
        c.each do |e|
          next if e.hidden?
          if e.class == Sketchup::Face
            faces.push(e)
          elsif e.class == Sketchup::Group
            next if e.description =~ /Ivy/
            faces.concat(wf.call(e.entities))
          elsif e.class == Sketchup::ComponentInstance
            #faces.concat(wf.call(e.entities))
          end
        end
        return faces
      end
      return wf.call(Sketchup.active_model.entities)
    end

    def self.all_faces_within(d, p)
      self.all_faces.select { |f|}
    end
  end
  module Face
    def self.is_on_face(face, p)
      Face.triangles(face).each do |v|
        return true if v.getBaryCentricCoordinates(p)
      end
      return false
    end


    def self.triangles(face)
      full_mesh = face.mesh(4)
      tris      = []
      i         = 1
      a         = full_mesh.polygons.map do |poly|
        poly = poly.map! do |vert|
          full_mesh.point_at(vert)
        end
        norm = full_mesh.normal_at(i)
        i    += 1
        tris.push(Geom::Triangle.new(poly, norm))

      end
      return tris
    end
  end
  class Triangle
    attr_reader :v0, :v1, :v2, :normal

    def initialize(vertices, norm)
      @normal = Geom::Vector3d.new([0, 0, 0])
      @v0     = vertices[0]
      @v1     = vertices[1]
      @v2     = vertices[2]
      @normal = norm
    end

    def getBaryCentricCoordinates(p)
      v0       = @v0.vector_to(@v2)
      v1       = @v0.vector_to(@v1)
      v2       = @v0.vector_to(p)

      dot00    = v0.dot(v0)
      dot01    = v0.dot(v1)
      dot02    = v0.dot(v2)
      dot11    = v1.dot(v1)
      dot12    = v1.dot(v2)

      invDenom = 1 / (dot00*dot11 - dot01*dot01)
      u        = (dot11*dot02-dot01*dot12) * invDenom
      v        = (dot00*dot12-dot01*dot02) * invDenom
      return (u>0)&& (v>0)&&(u+v<1)
    end

  end
  class Array
    def self.simplifyAngle(points, threshold)
      return points unless points.class == ::Array
      return points if points.length < 2
      pArray = points
      i      = 0
      for point in points
        next if i == points.length-1
        p = points[i-1]
        n = points[i+1]
        if p.vector_to(n).normalize.dot(point.vector_to(n).normalize) > threshold #|| point.distance(n) < (Si::SketchupIvy.nodeLength/(Si::Ivy.maxParents+2)) || point.distance(p) < (Si::SketchupIvy.nodeLength/(Si::Ivy.maxParents+2))
          pArray.delete_at(i)
        end
        i+=1
      end
      return pArray
    end

    def self.smoothAngle(points)
      return points unless points.class == ::Array
      return points unless points.length > 4

      pArray = points

      points.each_index do |i|
        next unless i > 0
        next unless points[i-1].class == Geom::Point3d
        next if i > points.length-3
        p0 = points[i-1]
        p1 = points[i]
        p2 = points[i+1]
        p3 = points[i+2]
        puts "#{p0.class}-#{p0}/#{p1.class}/#{p2.class}/#{p3.class}"
        puts "-------------------------"

        v1 = p1.vector_to(p0)
        v2 = p1.vector_to(p2)
        if (v1.dot(v2)).abs < 0.8
          puts "smoooooooooooooooooooooooth"
          port = Pdn::Interpolate::Catmull.independent_portion(p0, p1, p2, p3, 7)
          pArray.delete_at(i)
          pArray[i] = port
          pArray.flatten!
        end
      end
      return pArray
    end

    def self.simplifyDistance(points, threshold)
      return points unless points.class == ::Array
      return points unless points.length > 2
      pArray = points
      i      = 0
      for point in points
        next if i == points.length-1
        if i != 0 || i != points.length-1
          p = points[i-1]
          if point.distance(p) < threshold
            pArray.delete_at(i)
          end
          i+=1
        end
      end
      return pArray
    end
  end

  class Vector2d
    attr_reader :x, :y

    def initialize(x, y)
      @x, @y = x, y
    end

    def /(value)
      Vector2d.new(@x/value.to_f, @y/value.to_f)
    end

    def +(value)
      Vector2d.new(@x+value.x, @y+value.x)
    end

    def normalize
      if length > 0
        self./(length)
      else
        self
      end
    end

    def length
      Math::sqrt(@x**2+@y**2)
    end

    def to_polar
      phi = (@x == 0) ? 0 : Math::atan(@y/@x)

      if @x < 0
        phi +=3.14159
      else
        if @y <0
          phi += 2*3.14159
        end
      end
      return phi
    end

    def self.epsilon
      Vector2d.new(Float::EPSILON, Float::EPSILON)
    end
  end
  module Debug
    def self.benchmark(description, iterations, &block)
      puts "----------------------------------"
      puts description
      t1 = Time.now
      i  =0
      iterations.times do
        yield i
        i+=1
      end
      t2 = Time.now
      puts "#{t2-t1} seconds / #{(t2-t1)/iterations} sec.pr.iteration over #{iterations} iterations"
    end
  end
end
